"""The ingit runtime."""

import collections.abc
import functools
import logging
import pathlib
import platform
import re
import subprocess
import typing as t

from boilerplates.config import normalize_path
import git
import ordered_set

from .runtime_interface import ask
from .json_config import \
    json_to_file, default_machine_configuration, acquire_configuration, acquire_repos_configuration
from .repo_data import RepoData
from .project import Project

_LOG = logging.getLogger(__name__)


def regex_predicate(regex: str, name, tags, path, remotes):
    """Repo filtering function launched when "ingit -r 'regex' <command> ..." is used."""
    return (
        re.search(regex, name) is not None
        or any(re.search(regex, tag) is not None for tag in tags)
        or re.search(regex, str(path)) is not None
        or any(re.search(regex, name) for name, url in remotes.items()))


class Runtime:  # pylint: disable = too-many-instance-attributes
    """The ingit runtime."""

    def __init__(self, runtime_config_path: pathlib.Path, repos_config_path: pathlib.Path,
                 hostname: str | None = None, interactive: t.Optional[bool] = None):
        self.runtime_config_path = runtime_config_path
        self.repos_config_path = repos_config_path
        self._hostname = platform.node() if hostname is None else hostname
        self._interactive = interactive

        # prepare, possibly interactive
        self.runtime_config = acquire_configuration(self.runtime_config_path, 'runtime')
        self._machine_config = self._read_machine_config()

        self.repos_config = acquire_repos_configuration(self.repos_config_path)
        self.projects = self._read_projects()
        self.filtered_projects = list(self.projects)

    @property
    def repos_path(self):
        """Get path to where repositories are by default."""
        raw_path = self._machine_config['repos_path']
        if raw_path is None:
            return None
        return pathlib.Path(normalize_path(raw_path))

    @property
    def interactive(self):
        if self._interactive is None:
            return self._machine_config['interactive']
        return self._interactive

    def _read_machine_config(self):
        for machine in self.runtime_config['machines']:
            names = []
            assert ('name' in machine) ^ ('names' in machine), machine
            if 'name' in machine:
                names.append(machine['name'])
            else:
                names += machine['names']
            if '' in names or self._hostname in names:
                return machine

        ans = ask(f'No matching machine for "{self._hostname}" found in configuration. Add it?')
        if ans != 'y':
            raise ValueError(f'No matching machine for "{self._hostname}"'
                             f' found in configuration """{self.runtime_config}"""')
        machine = self.register_machine(self._hostname)
        return machine

    def _read_projects(self) -> t.Sequence[Project]:
        """Get list of all projects in repositories configuration."""
        projects = []
        for repo in self.repos_config['repos']:
            name = repo['name']
            assert not ('path' in repo and 'paths' in repo), repo
            raw_path = None
            if 'path' in repo:
                raw_path = repo['path']
            elif 'paths' in repo and self._hostname in repo['paths']:
                raw_path = repo['paths'][self._hostname]
            elif 'paths' in repo and '' in repo['paths']:
                raw_path = repo['paths']['']
            if raw_path is None:
                path = pathlib.Path(name)
            else:
                path = pathlib.Path(normalize_path(raw_path))
            if not path.is_absolute():
                if self.repos_path is None:
                    raise ValueError(f'configuration of repository "{name}" must contain absolute'
                                     ' path because repos_path in runtime configuration is not set')
                path = self.repos_path.joinpath(path)
            project = Project(name=name, tags=repo['tags'], path=path, remotes=repo['remotes'])
            projects.append(project)
        return projects

    def filter_projects(self, predicate: t.Union[collections.abc.Callable, str]):
        """Select subset of all of the projects registered projects for processing."""
        assert callable(predicate) or isinstance(predicate, str), type(predicate)
        if isinstance(predicate, str):
            predicate = functools.partial(regex_predicate, predicate)
        self.filtered_projects = [
            project for project in self.projects
            if predicate(project.name, project.tags, project.path, project.remotes)]

    def execute(self, command: str, **command_options):
        """Execute the runtime."""
        command_executor = {
            'summary': self.execute_ingit_command,
            'register': self.execute_ingit_command,
            'foreach': self.execute_ingit_command,
            'clone': self.execute_git_command,
            'init': self.execute_git_command,
            'fetch': self.execute_git_command,
            'checkout': self.execute_git_command,
            'merge': self.execute_git_command,
            'push': self.execute_git_command,
            'gc': self.execute_git_command,
            'status': self.execute_git_command}[command]

        command_executor(command, **command_options)

    def execute_ingit_command(self, command: str, **command_options):
        """Execute an ingit command, as opposed to a wrapper for a git built-in command."""
        command = {
            'summary': 'repositories_summary',
            'register': 'register_repository',
            'foreach': 'execute_command'}.get(command, command)
        implementation = getattr(self, command)
        implementation(**command_options)

    def execute_git_command(self, command: str, **command_options):
        """Execute a wrapper for a git built-in command."""
        command = {
            'gc': 'collect_garbage'}.get(command, command)
        for project in self.filtered_projects:
            implementation = getattr(project, command)
            try:
                implementation(**command_options)
            except RuntimeError:
                _LOG.exception('failed to execute command "%s" for project %s', command, project)

    def execute_command(self, cmd: str, timeout: t.Optional[int] = None):
        """Execute a command in each of the projects (after filtering was applied)."""
        for project in self.filtered_projects:
            if not project.path.is_dir():
                continue
            try:
                subprocess.run(cmd, check=True, timeout=timeout, shell=True, cwd=project.path)
            except subprocess.TimeoutExpired:
                _LOG.error(
                    'command "%s" in %s was aborted because it took too much time',
                    cmd, project)
            except subprocess.CalledProcessError as err:
                _LOG.error(
                    'command "%s" failed in %s:\nstderr: %s\nstdout: %s',
                    cmd, project, err.stderr, err.stdout)

    def repositories_summary(self):
        """Summarize registered repositories."""
        all_count = len(self.filtered_projects)
        was_filtered = all_count < len(self.repos_config['repos'])
        if was_filtered:
            print(f'Registered projects matching given conditions ({all_count}):')
        else:
            print(f'All registered projects ({all_count}):')
        initialised_count = 0
        for project in self.filtered_projects:
            if project.is_initialised:
                initialised_count += 1
            print(' -', project)

        if initialised_count == all_count:
            if all_count > 0:
                print('All of them are initialised.')
            else:
                print('')
        else:
            print(f'{initialised_count} of them are initialised'
                  f' ({all_count - initialised_count} not).')

        if self.repos_path is not None:
            self._unregistered_folders_summary()

    def _unregistered_folders_summary(self):
        assert self.repos_path is not None

        project_paths_in_root = ordered_set.OrderedSet()
        for project in self.projects:
            try:
                project_paths_in_root.add(project.path.relative_to(self.repos_path))
            except ValueError:
                pass

        unregistered_in_root, non_repo_paths_in_root = \
            self._find_unregistered_folders_in_root(project_paths_in_root)

        if unregistered_in_root:
            print(f'There are {len(unregistered_in_root)} unregistered git repositories'
                  f' in configured repositories root "{self.repos_path}".')
            for path in unregistered_in_root:
                print(path)

        if non_repo_paths_in_root:
            print(f'There are {len(non_repo_paths_in_root)} not versioned folders'
                  f' in configured repositories root "{self.repos_path}".')
            for path in non_repo_paths_in_root:
                print(path)

    def _find_unregistered_folders_in_root(self, project_paths_in_root: ordered_set.OrderedSet):
        assert self.repos_path is not None

        unregistered: ordered_set.OrderedSet[pathlib.Path] = ordered_set.OrderedSet()
        non_repo_paths: ordered_set.OrderedSet[pathlib.Path] = ordered_set.OrderedSet()
        for path in self.repos_path.iterdir():
            if not path.is_dir():
                continue
            try:
                _ = git.Repo(str(path))
            except git.InvalidGitRepositoryError:
                non_repo_paths.add(path)
                continue
            relative_path = path.relative_to(self.repos_path)
            if relative_path in project_paths_in_root:
                continue
            unregistered.add(path)
        return unregistered, non_repo_paths

    def register_machine(self, name: str) -> dict:
        """Add machine to ingit runtime configuration."""
        assert isinstance(name, str), type(name)
        assert name
        for machine in self.runtime_config['machines']:
            names = []
            if 'name' in machine:
                names.append(machine['name'])
            if 'names' in machine:
                names += machine['names']
            if '' in names or name in names:
                raise ValueError(f'machine "{name}" already in configuration')
        machine_config = default_machine_configuration(name)
        _LOG.warning('adding machine to configuration: %s', machine_config)
        self.runtime_config['machines'].append(machine_config)
        json_to_file(self.runtime_config, self.runtime_config_path)
        return machine_config

    def register_repository(self, path: pathlib.Path, tags: t.Sequence[str]):
        """Add repo to ingit repositories configuration."""
        repo = git.Repo(str(path))
        repo_data = RepoData(repo)
        repo_data.refresh()
        _LOG.debug('registering repository: %s', repo_data)
        repo_config = repo_data.generate_repo_configuration()

        self._postprocess_configured_repo_path(repo_config)

        if tags is not None:
            repo_config['tags'] += tags
        _LOG.warning('adding repo to configuration: %s', repo_config)
        index = None
        lowercase_name = repo_config['name'].lower()
        for i, project in enumerate(self.projects):
            lowercase_project_name = project.name.lower()
            if index is None and lowercase_project_name > lowercase_name:
                index = i
            if lowercase_project_name == lowercase_name:
                raise ValueError(
                    f'project named {project.name} already exists in current configuration')
        if index is None:
            index = len(self.projects)
        self.repos_config['repos'].insert(index, repo_config)
        json_to_file(self.repos_config, self.repos_config_path)

    def _postprocess_configured_repo_path(self, repo_config) -> None:
        try:
            absolute_repo_path = pathlib.Path(repo_config['path']).resolve()
        except ValueError:
            _LOG.info('cannot resolve repo path %s', repo_config['path'])
            return

        if self.repos_path is None:
            _LOG.warning(
                'repos path is not configured - registering absolute repo path "%s"',
                absolute_repo_path)
            repo_config['path'] = str(absolute_repo_path)
            return

        try:
            repo_path = absolute_repo_path.relative_to(self.repos_path)
        except ValueError:
            _LOG.warning(
                'resolved repo path "%s" is not within configured repos path "%s"'
                ' - registering absolute path', absolute_repo_path, self.repos_path)
            repo_config['path'] = str(absolute_repo_path)
            return

        if str(repo_path) == repo_config['name']:
            del repo_config['path']
            _LOG.warning(
                'resolved repo path "%s" is within configured repos path "%s" and resolves'
                ' to the repository name "%s" - registering without redundant path data',
                absolute_repo_path, self.repos_path, repo_path)
            return

        repo_config['path'] = str(repo_path)
        _LOG.warning(
            'resolved repo path "%s" is within configured repos path "%s" - registering'
            ' relative path "%s"', absolute_repo_path, self.repos_path, repo_path)



"""Entry point of the command-line interface of ingit package."""

# PYTHON_ARGCOMPLETE_OK

import logging

import boilerplates.logging

from .main import main


class Logging(boilerplates.logging.Logging):
    """Logging configuration."""

    packages = ['ingit']
    level_global = logging.INFO


if __name__ == '__main__':
    Logging.configure()
    main()



"""Human-readable information extraction from git.FetchInfo."""

import functools
import operator
import typing as t

import git

# list used by __create_fetch_info_strings()
_KNOWN_STRINGS = {
    git.FetchInfo.NEW_TAG: 'new tag',  # 1
    git.FetchInfo.NEW_HEAD: 'new branch',  # 2
    git.FetchInfo.HEAD_UPTODATE: 'already up to date',  # 4
    git.FetchInfo.TAG_UPDATE: 'tag update',  # 8
    git.FetchInfo.REJECTED: 'rejected',  # 16
    git.FetchInfo.FORCED_UPDATE: 'forced update',  # 32
    git.FetchInfo.FAST_FORWARD: 'can be fast-forwarded',  # 64
    git.FetchInfo.ERROR: 'error'  # 128
}

_KNOWN_FLAGS = functools.reduce(operator.or_, _KNOWN_STRINGS.keys())


def info_for_known_flags(
        flags: int, known_strings: t.Mapping[t.Any, str]) -> t.MutableSequence[str]:
    """Create string sequence that represents given flags according to given mapping."""
    info_strings = []
    for key, value in known_strings.items():
        if flags & key:
            info_strings.append(value)
    return info_strings


def info_for_unknown_flags(flags: int, known_flags: int) -> t.MutableSequence[str]:
    """Create string sequence that represents unknown flags according to given known flags mask."""
    info_strings = []
    unknown_flags = flags & ~(known_flags)
    if unknown_flags:
        info_strings.append(f'unknown flag(s): {unknown_flags} ({unknown_flags:032b})')
    return info_strings


def create_fetch_info_strings(info: git.FetchInfo):
    """Create FetchInfo strings."""
    info_strings = info_for_known_flags(info.flags, _KNOWN_STRINGS)
    prefix = '!!'
    if info.flags & info.HEAD_UPTODATE:
        prefix = '--'
    info_strings += info_for_unknown_flags(info.flags, _KNOWN_FLAGS)
    if info.note:
        info_strings.append(f'note: {info.note.strip()}')
    return (info_strings, prefix)



"""Tests for git operation progress reporting."""

import logging
import sys
import unittest

import boilerplates.logging

from ingit.action_progress import ActionProgress, _KNOWN_OPERATIONS_STRINGS as op_codes


class ActionProgressTests(unittest.TestCase):

    """Unit tests for ActionProgress."""

    def test_construct(self):
        """Can ActionProgress be constructed?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)

    def test_update(self):
        """Can ActionProgress be updated?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        print()

    def test_update_with_percentage(self):
        """Can ActionProgress be updated with percentage info?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        op_count = len(op_codes)
        op_index = 0
        print()
        for op_code in op_codes:
            op_index += 1
            apr.update(op_code, op_index, op_count)
            self.assertTrue(apr.printed_lines)
        apr.finalize()

    def test_finalize(self):
        """Can ActionProgress be finalised?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_update_and_finalize(self):
        """Can ActionProgress be finalised after updating several times?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_reuse(self):
        """Can ActionProgress instance be reused?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)

        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_reuse_after_one_step(self):
        """Can ActionProgress be reused after finalising after only one update?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
            apr.finalize()
            self.assertFalse(apr.printed_lines)

    def test_not_inline(self):
        """Can ActionProgress be ?"""
        apr = ActionProgress(inline=False)
        self.assertIsNotNone(apr)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_redirect_output(self):
        """Can ActionProgress output be redirected?"""
        apr = ActionProgress(f_d=sys.stderr)
        self.assertIsNotNone(apr)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_redirect_output_not_inline(self):
        """Can ActionProgress be used in non-inline mode? """
        log_stream = boilerplates.logging.StreamToCall(logging.debug)
        apr = ActionProgress(inline=False, f_d=log_stream)
        self.assertIsNotNone(apr)
        for op_code in op_codes:
            apr.update(op_code, 0, None)
            self.assertTrue(apr.printed_lines)
        log_stream.flush()
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_message(self):
        """Can ActionProgress be updated with custom message?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None, 'testu')
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_message_unusual(self):
        """Can ActionProgress be updated with message that begins with comma and space?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        apr.update(next(iter(op_codes)), 0, None, ', testu')
        self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_very_long_message(self):
        """Can ActionProgress be updated with message that overflows to next terminal line?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        print()
        for op_code in op_codes:
            apr.update(op_code, 0, None, 'testu' + ' testu' * 20)
            self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)

    def test_unknown_operation(self):
        """Can ActionProgress be updated with unknown operation code?"""
        apr = ActionProgress()
        self.assertIsNotNone(apr)
        self.assertFalse(apr.printed_lines)
        apr.update(2 ** 11, 0, None, ', this is status message of unknown operation')
        self.assertTrue(apr.printed_lines)
        apr.finalize()
        self.assertFalse(apr.printed_lines)



"""Interactive runtime interface of ingit."""

import collections.abc
import logging
import sys
import typing as t

import readchar

_INTERRUPTS = {chr(3)}

_NEWLINES = {'\n', '\r'}

OUT = logging.getLogger('ingit.interface.print')

OUT_HANDLER = logging.StreamHandler(sys.stdout)
OUT_HANDLER.setLevel(logging.NOTSET)
OUT.addHandler(OUT_HANDLER)


def default_template(question, answers, default):
    """Print a question with candidate answers, with the default marked accordingly.

    This is the default template for ask().
    """
    case_insensitive_answers = {_.lower() for _ in answers}
    case_insensitive = len(case_insensitive_answers) == len(answers)
    if case_insensitive:
        answers_print = [(a.upper() if a == default else a) for a in answers]
        return f'{question} [{"/".join(answers_print)}] '
    return f'{question} [{"/".join(answers)}] [default={default}] '


def ask(question: str, answers: t.Sequence[str] | None = None, default: t.Optional[str] = None,
        autoanswer: bool | str | None = None,
        template: collections.abc.Callable = default_template) -> str:
    """Ask a question that is to be answered with a single keystroke.

    By default, it is asked like this: "question [y/N] ".

    :param question: a proposition that is to be answered
    :param answers: list of valid answers, list of lower-case letters; if None, use ['y', 'n']
    :param default:
        a default answer, lower-case letter; if None, use last of the answer, 'n' by default
    :param autoanswer:
        answer to be automatically given; if True, the default answer will be given
    :param template: a function taking question and answers and returning string
    :return: a lower-case letter that is the answer to the question
    """
    assert isinstance(question, str), type(question)
    if answers is None:
        answers = ['y', 'n']
    assert isinstance(answers, collections.abc.Sequence), type(answers)
    assert answers, answers
    assert all(
        isinstance(_, str) and len(_) == 1 and _ not in _INTERRUPTS and _ not in _NEWLINES
        for _ in answers), answers
    assert len(answers) == len(set(answers)), answers
    if default is None:
        default = answers[-1]
    assert isinstance(default, str), type(default)
    assert default in answers, (default, answers)
    # if autoanswer is None and not RuntimeInterfaceConfig.interactive:
    #    autoanswer = True
    if autoanswer is True:
        autoanswer = default
    assert autoanswer is None or isinstance(autoanswer, str), type(autoanswer)
    assert autoanswer is None or autoanswer in answers, (autoanswer, answers)
    assert callable(template), type(template)
    print(template(question, answers, default), end='', flush=True)
    case_insensitive_answers = {_.lower() for _ in answers}
    case_insensitive = len(case_insensitive_answers) == len(answers)
    if case_insensitive:
        answers = list(case_insensitive_answers)
        default = default.lower()
        if autoanswer is not None:
            autoanswer = autoanswer.lower()
    answer = autoanswer
    while answer not in answers:
        answer = _wait_for_answer(answer, case_insensitive, default)
    print(answer)
    return str(answer)


def _wait_for_answer(answer: str | None, case_insensitive: bool, default: str) -> str:
    if answer is not None:
        print(answer, end='', flush=True)
    answer = readchar.readchar()
    assert answer is not None
    if case_insensitive:
        answer = answer.lower()
    if answer in _INTERRUPTS:
        raise KeyboardInterrupt()
    if answer in _NEWLINES:
        answer = default
    return answer



"""Unit tests for git-like commands of ingit."""

import collections
import logging
import os
import pathlib
import platform
import tempfile
import unittest
import unittest.mock

import git
import readchar

from ingit.main import main

_HERE = pathlib.Path(__file__).resolve().parent
_EXAMPLES_FOLDER = _HERE.joinpath('examples')

TEST_RUNTIME_CONFIG_PATH = _EXAMPLES_FOLDER.joinpath('runtime_config', 'example_initial.json')
TEST_REPOS_CONFIG_PATH = _EXAMPLES_FOLDER.joinpath('repos_config', 'example_initial.json')

PROJECT_NAMES = ('argunparse', 'transpyle', 'typed-astunparse')
PROJECT_NAME = PROJECT_NAMES[0]

_LOG = logging.getLogger(__name__)


def call_main(*args, answer='y'):
    with unittest.mock.patch.object(readchar, 'readchar', return_value=answer), \
            unittest.mock.patch.dict(os.environ, {
                'EXAMPLE_PROJECTS_PATH': os.environ.get('EXAMPLE_PROJECTS_PATH', '..')}):
        main(['--config', str(TEST_RUNTIME_CONFIG_PATH), '--repos', str(TEST_REPOS_CONFIG_PATH)]
             + list(args))


class Tests(unittest.TestCase):

    def setUp(self):
        assert 'INGIT_TEST_REPOS_PATH' not in os.environ
        self._tmpdir = tempfile.TemporaryDirectory()  # pylint: disable = consider-using-with
        os.environ['INGIT_TEST_REPOS_PATH'] = self._tmpdir.name
        _LOG.warning('set INGIT_TEST_REPOS_PATH="%s"', self._tmpdir.name)
        self.repos_path = pathlib.Path(self._tmpdir.name)

    def tearDown(self):
        if platform.system() != 'Windows':
            self._tmpdir.cleanup()
        del os.environ['INGIT_TEST_REPOS_PATH']

    def test_clone(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            self.assertFalse(repo_path.is_dir())
            call_main('-p', f'name == "{project_name}"', 'clone')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())

    def test_clone_no(self):
        repo_path = pathlib.Path(self.repos_path, PROJECT_NAME)
        call_main('-p', f'name == "{PROJECT_NAME}"', 'clone', answer='n')
        self.assertFalse(repo_path.is_dir())

    def test_clone_to_repo_dir(self):
        repo_path = pathlib.Path(self.repos_path, PROJECT_NAME)
        call_main('-p', f'name == "{PROJECT_NAME}"', 'clone')
        self.assertTrue(repo_path.is_dir())
        self.assertTrue(repo_path.joinpath('.git').is_dir())
        call_main('-r', f'^{PROJECT_NAME}$', 'clone')

    def test_clone_to_nonempty_dir(self):
        repo_path = pathlib.Path(self.repos_path, PROJECT_NAME)
        repo_path.mkdir()
        repo_path.joinpath('file.txt').touch()
        with self.assertLogs(
                logger=logging.getLogger('ingit.interface.print'), level=logging.CRITICAL):
            call_main('-p', f'name == "{PROJECT_NAME}"', 'clone')
        self.assertTrue(repo_path.is_dir())

    # def test_clone_with_multiple_remotes(self):

    def test_init(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            self.assertFalse(repo_path.is_dir())
            call_main('-p', f'name == "{project_name}"', 'init')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())

    def test_init_in_repo_dir(self):
        repo_path = pathlib.Path(self.repos_path, PROJECT_NAME)
        call_main('-p', f'name == "{PROJECT_NAME}"', 'clone')
        self.assertTrue(repo_path.is_dir())
        self.assertTrue(repo_path.joinpath('.git').is_dir())
        call_main('-p', f'name == "{PROJECT_NAME}"', 'init')

    def test_init_no(self):
        project_name = PROJECT_NAMES[0]
        call_main('-p', f'name == "{project_name}"', 'init')
        project_name = PROJECT_NAMES[1]
        repo_path = pathlib.Path(self.repos_path, project_name)
        call_main('-p', f'name == "{project_name}"', 'init', answer='n')
        self.assertFalse(repo_path.is_dir())

    def test_fetch(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'init')
            call_main('-p', f'name == "{project_name}"', 'fetch')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())

    def test_fetch_all(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'init')
            call_main('-p', f'name == "{project_name}"', 'fetch', '--all')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())

    def test_checkout_detached(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            self.assertFalse(repo_path.is_dir())
            call_main('-p', f'name == "{project_name}"', 'clone')
            self.assertTrue(repo_path.is_dir())
            repo = git.Repo(str(repo_path))
            head_ref = repo.head.commit.hexsha
            initial_commit = collections.deque(repo.iter_commits(), maxlen=1).pop()
            initial_ref = initial_commit.hexsha
            repo.git.checkout(initial_ref)
            call_main('-p', f'name == "{project_name}"', 'checkout', answer='n')
            self.assertEqual(repo.head.commit.hexsha, initial_ref)
            call_main('-p', f'name == "{project_name}"', 'checkout', answer='1')
            self.assertEqual(repo.head.commit.hexsha, head_ref)

    def test_checkout(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'clone')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())
            call_main('-p', f'name == "{project_name}"', 'checkout', answer='n')

    @unittest.expectedFailure
    def test_checkout_conflicting(self):
        """Checkout remote branch for which there already exists local branch having a different
        tracking branch, which will result in detached head.
        """
        # project_name = 'argunparse'
        self.fail()

    # def test_merge(self):
    #     for project_name in PROJECT_NAMES:
    #         repo_path = pathlib.Path(self.repos_path, project_name)
    #         call_main('-p', f'name == "{project_name}"', 'init')
    #         self.assertTrue(repo_path.is_dir())
    #         call_main('-p', f'name == "{project_name}"', 'merge')

    def test_push(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'init')
            self.assertTrue(repo_path.is_dir())
            call_main('-p', f'name == "{project_name}"', 'push')
            self.assertTrue(repo_path.is_dir())

    def test_gc(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'clone')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())
            call_main('-p', f'name == "{project_name}"', 'gc')

    def test_status(self):
        for project_name in PROJECT_NAMES:
            repo_path = pathlib.Path(self.repos_path, project_name)
            call_main('-p', f'name == "{project_name}"', 'clone')
            self.assertTrue(repo_path.is_dir())
            self.assertTrue(repo_path.joinpath('.git').is_dir())
            call_main('-p', f'name == "{project_name}"', 'status', answer='n')

    def test_skip_nonexisting(self):
        for command in ('fetch', 'checkout', 'merge', 'push', 'gc', 'status'):
            repo_path = pathlib.Path(self.repos_path, PROJECT_NAME)
            self.assertFalse(repo_path.is_dir())
            call_main('-p', f'name == "{PROJECT_NAME}"', command)
            self.assertFalse(repo_path.is_dir())



"""Human-readable information extraction from git.PushInfo."""

import functools
import operator

import git

from .fetch_flags import info_for_known_flags, info_for_unknown_flags

# used by __create_push_info_strings()
_KNOWN_STRINGS = {
    git.PushInfo.NEW_TAG: 'new tag',  # 1
    git.PushInfo.NEW_HEAD: 'new branch',  # 2
    git.PushInfo.NO_MATCH: 'no match',  # 4
    git.PushInfo.REJECTED: 'rejected',  # 8
    git.PushInfo.REMOTE_REJECTED: 'remote rejected',  # 16
    git.PushInfo.REMOTE_FAILURE: 'remote failure',  # 32
    git.PushInfo.DELETED: 'deleted',  # 64
    git.PushInfo.FORCED_UPDATE: 'forced update',  # 128
    git.PushInfo.FAST_FORWARD: 'was fast-forwarded',  # 256
    git.PushInfo.UP_TO_DATE: 'already up to date',  # 512
    git.PushInfo.ERROR: 'error'  # 1024
}

_KNOWN_FLAGS = functools.reduce(operator.or_, _KNOWN_STRINGS.keys())


def create_push_info_strings(info: git.PushInfo):
    """Create PushInfo strings."""
    info_strings = info_for_known_flags(info.flags, _KNOWN_STRINGS)
    prefix = '!!'
    if info.flags & info.UP_TO_DATE:
        prefix = '--'
    info_strings += info_for_unknown_flags(info.flags, _KNOWN_FLAGS)
    if info.summary:
        info_strings.append(f'summary: {info.summary.strip()}')
    return (info_strings, prefix)



"""Unit tests for ingit runtime."""

import logging
import os
import pathlib
import shutil
import tempfile
import unittest
import unittest.mock

from boilerplates.config import normalize_path
import readchar

from ingit.json_config import \
    RUNTIME_CONFIG_PATH, REPOS_CONFIG_PATH, REPO_LISTS_DIRECTORY_NAME, \
    default_runtime_configuration, default_repos_configuration, acquire_configuration, \
    acquire_repos_configuration

_HERE = pathlib.Path(__file__).resolve().parent
_EXAMPLES_FOLDER = _HERE.joinpath('examples')
_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_create_runtime_config(self):
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            with tempfile.NamedTemporaryFile() as tmp_file:
                path = pathlib.Path(tmp_file.name)
            with self.assertRaises(FileNotFoundError):
                acquire_configuration(path, 'runtime')

        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            with tempfile.NamedTemporaryFile() as tmp_file:
                path = pathlib.Path(tmp_file.name)
            runtime_config = acquire_configuration(path, 'runtime')
            path.unlink()
            self.assertEqual(runtime_config, default_runtime_configuration())

    def test_bad_runtime_config(self):
        bad_config_path = _EXAMPLES_FOLDER.joinpath('runtime_config', 'example_bad.json')
        with self.assertRaises(ValueError):
            acquire_configuration(bad_config_path, 'runtime')

    @unittest.skipUnless('CI' in os.environ, 'skipping test that affects user environment')
    def test_use_default_config_dir(self):
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            runtime_config = acquire_configuration(RUNTIME_CONFIG_PATH, 'runtime')
            runtime_config_path = normalize_path(RUNTIME_CONFIG_PATH)
            runtime_config_path.unlink()
            self.assertEqual(runtime_config, default_runtime_configuration())
            repos_config = acquire_configuration(REPOS_CONFIG_PATH, 'repos')
            repos_config_path = normalize_path(REPOS_CONFIG_PATH)
            repos_config_path.unlink()
            self.assertEqual(repos_config, default_repos_configuration())

    def test_create_repos_config(self):
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            with tempfile.NamedTemporaryFile() as tmp_file:
                path = pathlib.Path(tmp_file.name)
            with self.assertRaises(FileNotFoundError):
                acquire_configuration(path, 'repos')

        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            with tempfile.NamedTemporaryFile() as tmp_file:
                path = pathlib.Path(tmp_file.name)
            repos_config = acquire_configuration(path, 'repos')
            path.unlink()
            self.assertEqual(repos_config, default_repos_configuration())

    def test_create_repos_config_with_extra(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            with tempfile.NamedTemporaryFile(dir=tmp_dir) as tmp_file:
                path = pathlib.Path(tmp_file.name)
            _LOG.debug('temporary config path: %s', path)
            with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
                repos_config = acquire_repos_configuration(path)
            self.assertEqual(repos_config, default_repos_configuration())
            repo_lists_path = path.parent.joinpath(REPO_LISTS_DIRECTORY_NAME)
            _LOG.debug('repos lists path: %s', repo_lists_path)
            self.assertFalse(repo_lists_path.exists())
            repo_lists_path.mkdir()
            self.assertTrue(repo_lists_path.is_dir())
            repos_config = acquire_repos_configuration(path)
            self.assertEqual(repos_config, default_repos_configuration())
            extra_repos_config_path_source = _EXAMPLES_FOLDER.joinpath(
                'repos_config', 'repos.d', 'example_extra.json')
            extra_repos_config_path = repo_lists_path.joinpath('example_extra.json')
            shutil.copy(extra_repos_config_path_source, extra_repos_config_path)
            ignored_file_path = repo_lists_path.joinpath('ignored_file.txt')
            ignored_file_path.touch()
            repos_config = acquire_repos_configuration(path)
            self.assertNotEqual(repos_config, default_repos_configuration())
            self.assertIn('repos', repos_config)
            self.assertEqual(len(repos_config['repos']), 1)
            self.assertEqual(repos_config['repos'][0]['name'], 'nonexistent')



"""Single project."""

import collections
import collections.abc
import logging
import pathlib
import typing as t

from boilerplates.config import normalize_path
import git

from .repo_data import RepoData
from .action_progress import ActionProgress
from .runtime_interface import ask
from .fetch_flags import create_fetch_info_strings
from .push_flags import create_push_info_strings

_LOG = logging.getLogger(__name__)

OUT = logging.getLogger('ingit.interface.print')

_SPECIAL_REFS = {'HEAD', 'FETCH_HEAD'}


def normalize_url(url: str):
    return normalize_path(url)


def _print_log(
        ref_log, printed_header: str = '', head_count: int = 2, tail_count: int = 2) -> None:
    """Print log of commits occupying limited amount of space."""
    if printed_header:
        OUT.critical(printed_header)
    if len(ref_log) > head_count + tail_count + 1:
        for line in ref_log[:head_count]:
            OUT.critical(line)
        OUT.critical(
            '... skipped %i commits', len(ref_log) - head_count - tail_count)
        for line in ref_log[-tail_count:]:
            OUT.critical(line)
    else:
        for line in ref_log:
            OUT.critical(line)


class Project:
    """Single project."""

    def __init__(self, name: str, tags: t.Iterable[str], path: pathlib.Path,
                 remotes: t.Mapping[str, str]):
        assert isinstance(name, str), type(name)
        assert isinstance(tags, collections.abc.Iterable), type(tags)
        assert isinstance(path, pathlib.Path), type(path)
        assert isinstance(remotes, collections.abc.Mapping), type(remotes)
        self.name = name
        self.tags = set(tags)
        self.path = path
        self.remotes = collections.OrderedDict(list(remotes.items()))

        self.repo: t.Optional[RepoData] = None

    @property
    def is_existing(self) -> bool:
        """Return True if this project's working directory exists."""
        return self.path.is_dir()

    @property
    def has_git_folder(self) -> bool:
        """Return True if repo has .git folder."""
        return self.path.joinpath('.git').is_dir()

    @property
    def has_git_file(self) -> bool:
        """Return True if repo has .git file."""
        return self.path.joinpath('.git').is_file()

    @property
    def has_git_folder_or_file(self) -> bool:
        """Return True if repo has .git folder or it has .git file."""
        return self.has_git_folder or self.has_git_file

    @property
    def is_initialised(self) -> bool:
        """Return True if repo exists."""
        return self.is_existing and self.has_git_folder_or_file

    def link_repo(self):
        """Link this object with a repository existing in the filesystem."""
        assert self.repo is None, self.repo
        try:
            git_repo = git.Repo(normalize_path(str(self.path)))
        except (git.NoSuchPathError, git.InvalidGitRepositoryError) as err:
            raise RuntimeError(f'failed to link repo for path "{self.path}"') from err
        self.repo = RepoData(git_repo)

    def clone(self) -> None:
        """Execute "git clone --recursive --origin <remote-name> <remote-url> <path>".

        All of the <...> values are taken from project configuration.
        Values of <remote-...> are taken from default remote.

        This is followed by "git remote add <remote-name> <remote-url>" for all additional
        configured remotes.
        """
        if self.is_initialised:
            OUT.info('repo %s already initialised', self.path)
            return
        if self.is_existing:
            if any(self.path.iterdir()):
                OUT.critical(
                    '!! folder "%s" exists, and is not empty... please check it, and try again',
                    self.path)
                return
            OUT.info('folder %s exists, but is empty: continuing', self.path)

        remotes = list(self.remotes.items())
        if not remotes:
            raise ValueError(f'no configured remotes in repo {self.path}... cannot clone')
        (remote_name, remote_url), remaning_remotes = remotes[0], remotes[1:]

        if ask(f'Execute "git clone {remote_url} --recursive --origin={remote_name}'
               f' {self.path}"?') != 'y':
            OUT.warning('skipping %s', self.path)
            return

        progress = ActionProgress()
        try:
            self.repo = RepoData(git.Repo.clone_from(
                normalize_url(remote_url), normalize_path(str(self.path)), recursive=True,
                origin=remote_name, progress=progress.update))
        except git.GitCommandError as err:
            raise ValueError(f'error while cloning "{remote_url}" into "{self.path}"') from err
        progress.finalize()

        for remote_name, remote_url in remaning_remotes:
            self.repo.git.remote('add', remote_name, normalize_url(remote_url))
            self.repo.refresh()
            self._fetch_remote(remote_name)

    def init(self) -> None:
        """Execute "git init".

        This is followed by "git remote add <remote-name> <remote-url>" for each configured remote.
        """
        if self.is_initialised:
            OUT.error('repo %s already initialised', self.path)
            return
        if self.is_existing:
            raise ValueError('directory already exists... please check, delete it, and try again')

        if ask(f'Execute "git init {self.path}"?') != 'y':
            OUT.warning('skipping %s', self.path)
            return

        self.repo = RepoData(git.Repo.init(normalize_path(str(self.path))))

        for remote_name, remote_url in self.remotes.items():
            self.repo.git.remote('add', remote_name, normalize_url(remote_url))

    def fetch(self, all_remotes: bool = False) -> None:
        """Execute "git fetch --prune" on a remote of tracking branch of current branch.

        Or execute "git fetch --prune" for all remotes.
        """
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if self.repo is None:
            self.link_repo()
        assert self.repo is not None
        self.repo.refresh()

        if all_remotes:
            remote_names = self.repo.remotes
        else:
            remote_names = self._determine_remotes_to_fetch()

        self._fetch_remotes(*remote_names)

    def _determine_remotes_to_fetch(self):
        assert self.repo is not None
        if self.repo.active_branch is None:
            OUT.warning('!! "%s" is not on any branch, fetching all remotes', self.path)
            return self.repo.remotes
        try:
            remote_name, *_ = self.repo.tracking_branches[self.repo.active_branch]
        except KeyError:
            _LOG.warning('branch "%s" in "%s" not configured, fetching all remotes',
                         self.repo.active_branch, self.path)
            return self.repo.remotes
        except TypeError:
            OUT.warning('!! branch "%s" in "%s" has no tracking branch, fetching all remotes',
                        self.repo.active_branch, self.path)
            return self.repo.remotes
        return [remote_name]

    def _fetch_remote(self, remote_name: str) -> None:
        assert self.repo is not None
        fetch_infos: t.Sequence[git.FetchInfo]
        progress = ActionProgress()
        try:
            fetch_infos = self.repo.remotes[remote_name].fetch(prune=True, progress=progress)
        except git.GitCommandError as err:
            raise ValueError(f'error while fetching remote "{remote_name}"'
                             f' ("{self.repo.remotes[remote_name].url}") in "{self.name}"') from err
        progress.finalize()
        if not fetch_infos:
            _LOG.warning('no fetch info after fetching from remote "%s" in "%s"',
                         remote_name, self.name)
        for fetch_info in fetch_infos:
            self._print_fetch_info(fetch_info)
            if fetch_info.flags & fetch_info.FAST_FORWARD:
                # ans = ask('The fetch was fast-forward. Do you want to merge?')
                # if ans == 'y':
                #    raise NotImplementedError('merging not yet implemented')
                pass
        # return fetch_infos

    def _fetch_remotes(self, *remote_names: str) -> None:
        for remote_name in remote_names:
            self._fetch_remote(remote_name)

    def _print_fetch_info(self, fetch_info: git.FetchInfo) -> None:
        info_strings, prefix = create_fetch_info_strings(fetch_info)
        if not info_strings:
            return
        level = logging.WARNING if fetch_info.flags & git.FetchInfo.HEAD_UPTODATE \
            else logging.CRITICAL
        OUT.log(level, '%s fetched "%s" in "%s"; %s',
                prefix, fetch_info.ref, self.name, ', '.join(info_strings))

    def checkout(self) -> None:
        """Interactively select revision and execute "git checkout <revision>" on it.

        The list of branches to select from is composed by combining:
        - local branches
        - non-tracking branches on all remotes
        - local tags
        """
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if self.repo is None:
            self.link_repo()
        assert self.repo is not None
        self.repo.refresh()

        # if self.is_on_only_branch():
        #    return

        keys = r'1234567890abcdefghijklmopqrstuvwxz' \
            r'ABCDEFGHIJKLMOPRSTUVWXZ' \
            r''',.!?*&^%$#@;:'"/|\()[]<>{}-_+=~`'''
        revisions = self._prepare_checkout_list(keys)

        print('Checkout options:')
        for key, (revision, comment) in revisions.items():
            if comment is None:
                print(f'  {key}: {revision}')
            else:
                print(f'  {key}: {revision} ({comment})')

        answer = ask('Which branch to checkout?', answers=list(keys[:len(revisions)] + 'n'))
        if answer == 'n':
            return
        target, _ = revisions[answer]
        if target == self.repo.active_branch:
            return

        self.repo.git.checkout(target)

    def _prepare_checkout_list(self, keys: str):
        assert self.repo is not None
        revisions: t.Dict[str, t.Tuple[str, str | None]] = collections.OrderedDict()

        local_branches = list(self.repo.branches)
        remote_tracking_branches = set(
            _[:2] for _ in self.repo.tracking_branches.values() if _ is not None)
        remote_nontracking_branches = [
            (remote, branch) for remote, branch in self.repo.remote_branches.keys()
            if (remote, branch) not in remote_tracking_branches and branch not in _SPECIAL_REFS]
        local_tags = list(self.repo.tags)

        all_candidates_count: int = \
            len(local_branches) + len(remote_nontracking_branches) + len(local_tags)
        if all_candidates_count > len(keys):
            raise RuntimeError(
                'not enough available keys to create a single list of checkout candidates'
                f' - there are {len(keys)} keys ("{keys}") but {all_candidates_count} candidates:'
                f'\n- branches:{local_branches}, {remote_nontracking_branches}'
                f'\n- tags: {local_tags}')

        self._prepare_checkout_list_revisions(revisions, keys, remote_tracking_branches)

        if self.repo.active_branch is None:
            revisions['n'] = ('---', 'keep no branch/tag')
        else:
            # assert revisions['1'][0] == active_branch
            for key, (revision, comment) in revisions.items():
                if revision == self.repo.active_branch:
                    _ = 'no change'
                    comment = f'{comment}, {_}' if comment else _
                    revisions[key] = (revision, comment)
                    break

        return revisions

    def _prepare_checkout_list_revisions(self, revisions, keys, remote_tracking_branches) -> None:
        assert self.repo is not None
        index = 0
        for branch in self.repo.branches:
            revisions[keys[index]] = (branch, None)
            index += 1

        for tag in self.repo.tags:
            revisions[keys[index]] = (tag.name, 'tag')
            index += 1

        for (remote, branch) in self.repo.remote_branches.keys():
            if (remote, branch) in remote_tracking_branches \
                    or branch in _SPECIAL_REFS:
                continue
            if branch in self.repo.branches:
                to_checkout = f'{remote}/{branch}'
            else:
                to_checkout = branch
            revisions[keys[index]] = (to_checkout, f'based on {remote}')
            index += 1

    def merge(self) -> None:
        """Execute "git merge" for current branch."""
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if self.repo is None:
            self.link_repo()
        raise NotImplementedError('merging not yet implemented')

    def push(self, all_branches: bool = False) -> None:
        """Push current branch to it's tracking branch.

        Or, push all local branches to their tracking branches.
        """
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if self.repo is None:
            self.link_repo()
        assert self.repo is not None

        branches_to_push = []  # type: t.List[str]

        if all_branches:
            raise NotImplementedError('pushing not yet implemented')

        if self.repo.active_branch is None:
            OUT.warning('not pushing "%s" because there is no active branch...', self.path)
            return

        branches_to_push.append(self.repo.active_branch)

        pushed_branches_per_remote = {}  # type: t.Dict[str, t.Dict[str, t.Optional[str]]]
        for local_branch in branches_to_push:
            tracking_branch = self.repo.tracking_branches[local_branch]
            if tracking_branch is None:
                remote = next(iter(self.repo.remotes.keys()))
                OUT.warning('pushing "%s" to "%s"...', self.path, remote)
                remote_branch = None
            else:
                remote, remote_branch, _ = tracking_branch
            if remote in pushed_branches_per_remote:
                pushed_branches_per_remote[remote][local_branch] = remote_branch
            else:
                pushed_branches_per_remote[remote] = {local_branch: remote_branch}

        for remote_name, branch_mapping in pushed_branches_per_remote.items():
            push_infos = self._push_single_remote(remote_name, branch_mapping)
            for push_info in push_infos:
                self._print_push_info(push_info)

    def _push_single_remote(
            self, remote_name: str,
            branch_mapping: t.Mapping[str, t.Optional[str]]) -> t.Sequence[git.PushInfo]:
        assert self.repo is not None
        push_refspec = [f'{local_branch}' if remote_branch is None
                        else f'{local_branch}:{remote_branch}'
                        for local_branch, remote_branch in branch_mapping.items()]
        _LOG.warning('push refspec: %s', push_refspec)
        progress = ActionProgress()
        try:
            push_infos = self.repo.remotes[remote_name].push(
                refspec=push_refspec, with_extended_output=True, progress=progress)
        except git.GitCommandError as err:
            raise ValueError(
                f'error while pushing branches {branch_mapping} to remote "{remote_name}"'
                f' ("{self.repo.remotes[remote_name].url}") in "{self.name}"') from err
        progress.finalize()
        return push_infos

    def _print_push_info(self, push_info: git.PushInfo) -> None:
        info_strings, prefix = create_push_info_strings(push_info)
        if not info_strings:
            return
        level = logging.WARNING if push_info.flags & git.PushInfo.UP_TO_DATE \
            else logging.CRITICAL
        OUT.log(level, '%s pushed "%s" to "%s" in "%s"; %s', prefix, push_info.local_ref,
                push_info.remote_ref, self.name, ', '.join(info_strings))

    def collect_garbage(self) -> None:
        """Execute "git gc --agressive --prune"."""
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if self.repo is None:
            self.link_repo()
        assert self.repo is not None

        try:
            self.repo.git.gc(aggressive=True, prune=True)
        except git.GitCommandError as err:
            raise ValueError(f'error while collecting garbage in "{self.path}"') from err

    def status(self, ignored: bool = False) -> None:
        """Execute "git status --short" and run "git gui" if there is any output.

        The "ignored" flag is equivalent to adding "--ignored".
        """
        if not self.is_existing:
            OUT.info('skipping non-existing "%s"...', self.path)
            return
        if not self.is_initialised and not any(self.path.iterdir()):
            OUT.critical('!! directory "%s" exists but is empty', self.path)
            return
        if self.repo is None:
            self.link_repo()
        assert self.repo is not None

        try:
            status_log = self.repo.git.status(short=True, branch=True, ignored=ignored).splitlines()
        except git.GitCommandError as err:
            raise ValueError(f'error while getting status of "{self.path}"') from err

        if len(status_log) > 1:
            OUT.critical('!! unclear status in "%s":', self.path)
            for line in status_log:
                OUT.critical(line)

        self.repo.refresh()
        self._status_branch()
        self._status_remotes()

    def _get_log(self, start_ref: str, end_ref: str) -> str:
        """Get log as if start_ref..end_ref was used."""
        assert self.repo is not None
        refs = f'{start_ref}..{end_ref}'
        try:
            git_log = self.repo.git.log('--color=always', '--pretty=oneline', refs)
        except git.GitCommandError as err:
            raise RuntimeError(f'error while getting log for "{refs}" of "{self.path}"') from err
        return git_log.splitlines()

    def _status_branch(self, branch: t.Optional[str] = None) -> None:
        """Evaluate the status of single branch by comparing it to the remote branch.

        auto-answers used in this function: create_remote_branch, push, merge, forget_locally
        """
        assert self.repo is not None
        if branch is None:
            branch = self.repo.active_branch
        if branch is None:
            OUT.warning('cannot diagnose branch status in "%s" -- not on any branch', self.path)
            return
        tracking_branch_data = self.repo.tracking_branches[branch]
        if tracking_branch_data is None:
            OUT.warning('cannot diagnose branch status in "%s"'
                        ' -- current branch has no tracking branch', self.path)
            return
        tracking_branch = '/'.join(tracking_branch_data[:2])
        try:
            self.repo.git.rev_parse(tracking_branch)
        except git.GitCommandError:
            OUT.warning(
                'cannot diagnose branch status in "%s"'
                ' -- tracking branch of the current branch does not exist', self.path)
            return
        not_pushed_log = self._get_log(tracking_branch, branch)
        if not_pushed_log:
            _print_log(
                not_pushed_log,
                f'!! not pushed commits from "{branch}" to "{tracking_branch}" in "{self.path}":')
            # self.push_single_remote(
            #    tracking_branch_data[0], ['{0}:{0}'.for.mat(branch, tracking_branch_data[1])])
        not_merged_log = self._get_log(branch, tracking_branch)
        if not_merged_log:
            _print_log(
                not_merged_log,
                f'!! not merged commits from "{tracking_branch}" to "{branch}" in "{self.path}":')
            # answer = self.interface.get_answer('merge')
            # if answer in ['y', 'm']:
            #     self.merge_single_branch(remote, branch)
            # elif answer == 'b':
            #     self.rebase_single_branch(remote, branch)
            # elif answer == 'r':
            #     self.hard_reset_single_branch(remote, branch)
            # elif self.interface.confirm('forget_locally'):
            #     self.repo.git.update_ref(f'refs/remotes/{remote}/{branch}', branch)

    def _status_remotes(self):
        assert self.repo is not None
        assert all(len(list(v.urls)) == 1 for v in self.repo.remotes.values()), self.repo.remotes
        remote_names_in_config = set(self.remotes)
        remote_names = set(self.repo.remotes)
        remotes_in_config = {k: v.replace('\\', '/') for k, v in self.remotes.items()}
        remotes = {k: tuple(v.urls)[0].replace('\\', '/') for k, v in self.repo.remotes.items()}

        extra_remote_names = remote_names - remote_names_in_config
        missing_remote_names = remote_names_in_config - remote_names
        extra_remotes = dict(set(remotes.items()) - set(remotes_in_config.items()))
        missing_remotes = dict(set(remotes_in_config.items()) - set(remotes.items()))
        if not extra_remotes and not missing_remotes:
            return
        OUT.critical('!! repo "%s" has different remotes than it should', self.path)

        for name, url in extra_remotes.items():
            self._status_handle_extra_remote(
                name, url, extra_remote_names, remotes_in_config, remotes, missing_remotes)

        for name, url in missing_remotes.items():
            OUT.critical('!! missing remote: "%s"', name)
            if name not in missing_remote_names:
                OUT.critical('!! misconfigured remote: "%s"', name)
                continue
            ans = ask(f'Add remote "{name}" with url "{self.remotes[name]}"?')
            if ans == 'y':
                self.repo.git.remote('add', name, self.remotes[name])

    def _status_handle_extra_remote(
            # pylint: disable = too-many-arguments, too-many-positional-arguments
            self, name: str, url: str, extra_remote_names, remotes_in_config, remotes,
            missing_remotes) -> None:
        assert self.repo is not None
        if url in missing_remotes.values():
            OUT.critical('!! renamed remote: "%s"', name)
            new_name = [k for k, v in missing_remotes.items() if url == v][0]
            ans = ask(f'Rename remote from "{name}" to "{new_name}"?')
            if ans != 'y':
                return
            self.repo.git.remote('rename', name, new_name)
            del missing_remotes[new_name]
            return
        if name in missing_remotes.keys():
            OUT.critical('!! url changed for remote: "%s"', name)
            ans = ask(f'Change URL of remote "{name}" from "{url}" to "{remotes_in_config[name]}"?')
            if ans != 'y':
                return
            self.repo.git.remote('set-url', name, remotes_in_config[name])
            del missing_remotes[name]
            return

        OUT.critical('!! extra remote: "%s"', name)
        if name not in extra_remote_names:
            OUT.critical('!! misconfigured remote: "%s"', name)
            return
        ans = ask(f'Remove remote "{name}" with url "{remotes[name]}"?')
        if ans != 'y':
            return
        self.repo.git.remote('remove', name)

    def __str__(self):
        fields = [f'path="{self.path}"']
        if not self.is_initialised:
            if not self.is_existing:
                fields.append('not existing')
            else:
                fields.append('not initialized')
        if self.tags:
            fields.append(f'tags={self.tags}')
        return f'{self.name} ({", ".join(fields)})'



"""Initialization of tests for ingit package."""

import logging

from ingit.__main__ import Logging


class TestsLogging(Logging):
    """Test logging configuration."""

    level_global = logging.DEBUG


TestsLogging.configure()



"""Unit tests for ingit command-line interface."""

import contextlib
import os
import unittest
import unittest.mock

from boilerplates.config import normalize_path
from boilerplates.packaging_tests import run_module
import readchar

from ingit.json_config import RUNTIME_CONFIG_PATH, REPOS_CONFIG_PATH
from ingit.main import main


class Tests(unittest.TestCase):

    def test_script(self):
        with self.assertRaises(SystemExit):
            run_module('ingit')
        run_module('ingit', run_name='not_main')

    def test_help(self):
        with open(os.devnull, 'a', encoding='utf-8') as devnull:
            for flags in (['-h'], ['--help']):
                with self.assertRaises(SystemExit):
                    with contextlib.redirect_stdout(devnull):
                        main(flags)

    def test_filtered_register(self):
        with self.assertRaises(SystemExit):
            main(['-p', 'something', 'register'])
        with self.assertRaises(SystemExit):
            main(['-r', 'True', 'register'])

    @unittest.skipUnless('CI' in os.environ, 'skipping test that affects user environment')
    def test_create_configs(self):
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            runtime_config_path = normalize_path(RUNTIME_CONFIG_PATH)
            repos_config_path = normalize_path(REPOS_CONFIG_PATH)
            self.assertFalse(runtime_config_path.exists())
            self.assertFalse(repos_config_path.exists())
            main(['register'])
            self.assertTrue(runtime_config_path.exists())
            self.assertTrue(repos_config_path.exists())
            runtime_config_path.unlink()
            repos_config_path.unlink()



"""Unit tests for ingit runtime interface."""

import logging
import unittest
import unittest.mock

import readchar

from ingit.runtime_interface import ask

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def test_ask(self):
        for answer in ('n', 'y'):
            with unittest.mock.patch.object(readchar, 'readchar', return_value=answer):
                actual_answer = ask('Really?')
                self.assertEqual(actual_answer, answer)

    def test_ask_default_answer(self):
        for keypress in ('\n', '\r'):
            with unittest.mock.patch.object(readchar, 'readchar', return_value=keypress):
                actual_answer = ask('Really?')
                self.assertEqual(actual_answer, 'n')
            for deafult_answer in ('n', 'y'):
                with unittest.mock.patch.object(readchar, 'readchar', return_value=keypress):
                    actual_answer = ask('Really?', default=deafult_answer)
                    self.assertEqual(actual_answer, deafult_answer)

    def test_ask_keyboard_interrupt(self):
        with self.assertRaises(KeyboardInterrupt):
            with unittest.mock.patch.object(readchar, 'readchar', return_value=chr(3)):  # Ctrl+C
                ask('Really?')



"""Unit tests for operations on a single project."""

import contextlib
import logging
import os
import pathlib
import tempfile
import unittest
import unittest.mock

import boilerplates.git_repo_tests
import git
import readchar

from ingit.project import Project

_HERE = pathlib.Path(__file__).resolve().parent

_REMOTE = str(_HERE.parent.joinpath('.git'))

_LOG = logging.getLogger(__name__)


class Tests(boilerplates.git_repo_tests.GitRepoTests):
    # pylint: disable = too-many-public-methods

    def setUp(self):
        super().setUp()
        self.default_branch_name = git.GitConfigParser(
            read_only=True).get_value('init', 'defaultBranch', default='master')

    def test_example(self):
        project = Project('example', ['tag1', 'tag2'], self.repo_path, {})
        self.assertEqual(project.name, 'example')
        self.assertSetEqual(project.tags, {'tag1', 'tag2'})

    def test_clone(self):
        project = Project('example', [], self.repo_path.joinpath('example'), {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            project.clone()
        self.assertTrue(self.repo_path.joinpath('example', '.git').is_dir())

    def test_clone_to_existing_empty_dir(self):
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        self.assertTrue(self.repo_path.exists())
        self.assertFalse(self.repo_path.joinpath('.git').exists())
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            project.clone()
        self.assertTrue(self.repo_path.joinpath('.git').is_dir())

    def test_clone_initialized(self):
        self.git_init()
        project = Project('example', [], self.repo_path, {})
        project.clone()

    def test_clone_no_remote(self):
        project = Project('example', [], self.repo_path.joinpath('example'), {})
        with self.assertRaises(ValueError):
            project.clone()

    def test_clone_nonexisting(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            nonexisting_path = pathlib.Path(temp_dir)
        with tempfile.TemporaryDirectory() as temp_dir:
            path = pathlib.Path(temp_dir)
            project = Project('example', [], path, {'origin': str(nonexisting_path)})
            self.assertTrue(path.is_dir())
            with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
                with self.assertRaises(ValueError):
                    project.clone()

    def test_clone_no(self):
        project = Project('example', [], self.repo_path.joinpath('example'), {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.clone()
        self.assertFalse(self.repo_path.joinpath('example', '.git').is_dir())

    def test_init(self):
        project = Project('example', [], self.repo_path.joinpath('example'), {})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            project.init()
        self.assertTrue(self.repo_path.joinpath('example', '.git').is_dir())

    def test_init_existing(self):
        project = Project('example', [], self.repo_path, {})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            with self.assertRaises(ValueError):
                project.init()

    def test_init_initialized(self):
        self.git_init()
        project = Project('example', [], self.repo_path, {})
        project.init()

    def test_init_no(self):
        project = Project('example', [], self.repo_path.joinpath('example'), {})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.init()
        self.assertFalse(self.repo_path.joinpath('example', '.git').is_dir())

    def test_link_repo_invalid(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = pathlib.Path(temp_dir)
            self.assertTrue(path.is_dir())
            project = Project('example', [], path, {})
            with self.assertRaises(RuntimeError):
                project.link_repo()
        self.assertFalse(path.exists())
        project = Project('example', [], path, {})
        with self.assertRaises(RuntimeError):
            project.link_repo()

    # def test_fetch(self):
    #    pass

    # def test_fetch_all(self):
    #    pass

    # def test_checkout(self):
    #    pass

    def test_checkout_case_sensitive_no(self):
        self.git_init()
        self.git_commit_new_file()
        for i in range(0, 88):
            self.repo.git.branch(f'branch_{i:02}')
        project = Project('example', [], self.repo_path, {})
        project.link_repo()
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.checkout()

        def readchar_once():
            readchar_once.called_count += 1
            if readchar_once.called_count > 1:
                self.assertLessEqual(readchar_once.called_count, 2)
                raise TimeoutError
            return 'N'
        readchar_once.called_count = 0
        with unittest.mock.patch.object(readchar, 'readchar', readchar_once):
            with self.assertRaises(TimeoutError):
                project.checkout()

    def test_checkout_too_many_branches(self):
        self.git_init()
        self.git_commit_new_file()
        project = Project('example', [], self.repo_path, {})
        project.link_repo()
        for i in range(0, 88):
            self.repo.git.branch(f'branch_{i:02}')
            project.repo.refresh()
            with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
                with open(os.devnull, 'a', encoding='utf-8') as devnull:
                    with contextlib.redirect_stdout(devnull):
                        project.checkout()
        self.repo.git.branch('devel')
        project.repo.refresh()
        with self.assertRaises(RuntimeError):
            with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
                project.checkout()

    # @unittest.expectedFailure
    # def test_merge(self):
    #    pass

    def test_push(self):
        separate_repo = boilerplates.git_repo_tests.GitRepoTests()
        separate_repo.setUp()
        separate_repo.git_init()
        separate_repo.git_commit_new_file()
        separate_repo.repo.git.checkout('-b', 'devel')
        separate_repo.repo.git.branch('-d', self.default_branch_name)
        target_url = str(separate_repo.repo_path)

        self.git_init()
        self.repo.git.remote('add', 'target', target_url)
        project = Project('example', [], self.repo_path, {'target': target_url})
        project.push()
        # project.push(all_branches=True)

        self.git_commit_new_file()
        self.repo.git.push('target', self.default_branch_name)
        self.git_commit_new_file()
        project.repo.refresh()
        self.assertEqual(project.repo.active_branch, self.default_branch_name)
        self.assertIsNone(project.repo.tracking_branches[self.default_branch_name])
        project.push()
        # project.push(all_branches=True)

        self.repo.git.branch(
            self.default_branch_name, set_upstream_to=f'target/{self.default_branch_name}')
        project.repo.refresh()
        self.assertEqual(project.repo.active_branch, self.default_branch_name)
        self.assertTupleEqual(
            project.repo.tracking_branches[self.default_branch_name][:2],
            ('target', self.default_branch_name))
        project.push()
        # project.push(all_branches=True)

        project.repo.refresh()

        separate_repo.tearDown()

    # def test_gc(self):
    #    pass

    def test_status(self):
        self.git_init()
        self.repo.git.remote('add', 'origin', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        project.status()
        project.status(ignored=True)

    def test_status_unclear(self):
        self.git_init()
        self.repo.git.remote('add', 'origin', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        path = self.git_commit_new_file()
        self.git_modify_file(path)
        project.status()
        self.git_modify_file(path, add=True)
        project.status()
        self.git_modify_file(path, commit=True)
        project.status()
        project.status(ignored=True)

    def test_status_not_pushed(self):
        self.git_clone('origin', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        self.git_commit_new_file()
        project.status()

    def test_status_not_pushed_many(self):
        self.git_clone('origin', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        for _ in range(10 + 10 + 1 + 1):
            self.git_commit_new_file()
        project.status()

    def test_status_not_merged(self):
        self.git_clone('origin', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        self.repo.git.reset('--hard', 'HEAD~1')
        project.status()

    def test_status_extra_remote(self):
        self.git_init()
        self.repo.git.remote('add', 'origin', _REMOTE)
        self.repo.git.remote('add', 'mirror', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.status()

    def test_status_missing_remote(self):
        self.git_init()
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            project.status()
        self.assertIn('origin', self.repo.git.remote(v=True))
        self.assertIn(_REMOTE, self.repo.git.remote(v=True))

    def test_status_missing_remote_no(self):
        self.git_init()
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.status()
        self.assertNotIn('origin', self.repo.git.remote(v=True))
        self.assertNotIn(_REMOTE, self.repo.git.remote(v=True))

    def test_status_tracking_branch_gone(self):
        # create 1st repo
        self.git_init()
        self.git_commit_new_file()
        self.repo.git.checkout('-b', 'temporary_branch')
        self.git_commit_new_file()
        with tempfile.TemporaryDirectory() as temp_dir:
            # create 2nd repo by cloning the 1st
            path = pathlib.Path(temp_dir)
            cloned_repo = git.Repo.clone_from(self.repo_path, path, origin='source')
            project = Project('example', [], path, {'source': str(self.repo_path)})
            _LOG.debug('linking repo at %s', path)
            project.link_repo()
            project.status()
            # delete the branch in the 1st repo
            self.repo.git.checkout(self.default_branch_name)
            self.repo.git.branch('-D', 'temporary_branch')
            # fetch the changes
            cloned_repo.git.fetch(prune=True)
            _LOG.debug('refreshing repo at %s', path)
            project.repo.refresh()
            output = cloned_repo.git.branch('-vav')
            _LOG.debug('git branch -vav output:\n%s', output)
            self.assertIn('[source/temporary_branch: gone]', output)
            project.status()

    def test_status_bad_remote(self):
        self.git_init()
        self.repo.git.remote('add', 'mirror', _REMOTE)
        project = Project('example', [], self.repo_path, {'origin': _REMOTE})
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            project.status()
        self.assertNotIn('origin', self.repo.git.remote(v=True))
        self.assertIn(_REMOTE, self.repo.git.remote(v=True))



"""Unit tests for git repository data gathering."""

import logging
import tempfile

import boilerplates.git_repo_tests
import git

from ingit.repo_data import RepoData

_LOG = logging.getLogger(__name__)


class Tests(boilerplates.git_repo_tests.GitRepoTests):

    def setUp(self):
        super().setUp()
        self.default_branch_name = git.GitConfigParser(
            read_only=True).get_value('init', 'defaultBranch', default='master')

    def test_empty_repo(self):
        self.git_init()
        repo = RepoData(self.repo)
        repo.refresh()
        self.assertDictEqual(repo.remotes, {})
        self.assertIsNone(repo.default_remote)
        self.assertDictEqual(repo.branches, {})
        self.assertIsNone(repo.active_branch)
        self.assertIsNone(repo.current_tracking_branch)
        self.assertDictEqual(repo.remote_branches, {})
        self.assertDictEqual(repo.tracking_branches, {})

    def test_detached_head(self):
        self.git_init()
        self.git_commit_new_file()
        self.repo.git.checkout(self.repo_head_hexsha)
        repo = RepoData(self.repo)
        repo.refresh()
        self.assertDictEqual(repo.remotes, {})
        self.assertIsNone(repo.default_remote)
        self.assertSetEqual(set(repo.branches), {self.default_branch_name})
        self.assertIsNone(repo.active_branch)
        self.assertIsNone(repo.current_tracking_branch)
        self.assertDictEqual(repo.remote_branches, {})
        self.assertSetEqual(set(repo.tracking_branches), {self.default_branch_name})

    def test_no_remotes(self):
        self.git_init()
        self.git_commit_new_file()
        repo = RepoData(self.repo)
        repo.refresh()
        self.assertDictEqual(repo.remotes, {})
        self.assertIsNone(repo.default_remote)
        self.assertSetEqual(set(repo.branches), {self.default_branch_name})
        self.assertEqual(repo.active_branch, self.default_branch_name)
        self.assertIsNone(repo.current_tracking_branch)
        self.assertDictEqual(repo.remote_branches, {})
        self.assertSetEqual(set(repo.tracking_branches), {self.default_branch_name})

    def test_has_remotes(self):
        self.git_init()
        self.git_commit_new_file()
        origin_name = 'my-origin'
        with tempfile.TemporaryDirectory() as tmpdir:
            git_repo = git.Repo.clone_from(self.repo_path, tmpdir, origin=origin_name)
            repo = RepoData(git_repo)
            repo.refresh()
            self.assertGreater(len(repo.remotes), 0)
            self.assertEqual(repo.default_remote, origin_name)
            self.assertSetEqual(set(repo.branches), {self.default_branch_name})
            self.assertEqual(repo.active_branch, self.default_branch_name)
            self.assertEqual(
                repo.current_tracking_branch, f'{origin_name}/{self.default_branch_name}')
            self.assertGreater(len(repo.remote_branches), 0)
            self.assertEqual(len(repo.tracking_branches), 1)
            key, value = next(iter(repo.tracking_branches.items()))
            self.assertEqual(key, self.default_branch_name)
            self.assertEqual(value.remote_name, origin_name)
            self.assertEqual(value.tracking_branch_name, self.default_branch_name)



"""Unit tests for ingit runtime."""

import logging
import pathlib
import platform
import shutil
import tempfile
import unittest
import unittest.mock

import readchar

from ingit.runtime import Runtime

_HERE = pathlib.Path(__file__).resolve().parent
_EXAMPLES_FOLDER = _HERE.joinpath('examples')

TEST_RUNTIME_CONFIG_PATH = _EXAMPLES_FOLDER.joinpath('runtime_config', 'example_initial.json')
TEST_REPOS_CONFIG_PATH = _EXAMPLES_FOLDER.joinpath('repos_config', 'example_paths.json')

_LOG = logging.getLogger(__name__)


class Tests(unittest.TestCase):

    def setUp(self):
        with tempfile.NamedTemporaryFile() as tmp_file1:
            with tempfile.NamedTemporaryFile() as tmp_file2:
                self.repos_config_path = pathlib.Path(tmp_file2.name)
            self.runtime_config_path = pathlib.Path(tmp_file1.name)

    def tearDown(self):
        if self.runtime_config_path.is_file():
            self.runtime_config_path.unlink()
        if self.repos_config_path.is_file():
            self.repos_config_path.unlink()

    def test_nothing(self):
        pass

    def test_create_configs(self):
        self.assertFalse(self.runtime_config_path.is_file())
        self.assertFalse(self.repos_config_path.is_file())
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            runtime = Runtime(self.runtime_config_path, self.repos_config_path)
        self.assertEqual(self.runtime_config_path, runtime.runtime_config_path)
        self.assertEqual(self.repos_config_path, runtime.repos_config_path)
        self.assertTrue(self.runtime_config_path.is_file())
        self.assertTrue(self.repos_config_path.is_file())

    def test_register_machine(self):
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            runtime = Runtime(self.runtime_config_path, self.repos_config_path)
        runtime.register_machine('blah')
        self.assertIn('blah', [machine['name'] for machine in runtime.runtime_config['machines']])
        with self.assertRaises(ValueError):
            runtime.register_machine(platform.node())

    def test_autoadd_machine(self):
        """Test adding machine into an existing config."""
        runtime_config_path = _EXAMPLES_FOLDER.joinpath('runtime_config', 'example_names.json')
        shutil.copy(str(runtime_config_path), str(self.runtime_config_path))
        with unittest.mock.patch.object(readchar, 'readchar', return_value='n'):
            with self.assertRaises(ValueError):
                Runtime(self.runtime_config_path, self.repos_config_path)
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            runtime = Runtime(self.runtime_config_path, self.repos_config_path)
        self.assertIn(platform.node(), [
            machine.get('name') for machine in runtime.runtime_config['machines']])

    def test_path_depends_on_machine(self):
        shutil.copy(str(TEST_RUNTIME_CONFIG_PATH), str(self.runtime_config_path))
        shutil.copy(str(TEST_REPOS_CONFIG_PATH), str(self.repos_config_path))

        for hostname, path in [
                (f'example_machine{i}', pathlib.Path(f'/example_path_{i}'))
                for i in range(1, 4)]:
            runtime = Runtime(self.runtime_config_path, self.repos_config_path, hostname=hostname)
            runtime.filter_projects(lambda name, *_: name == 'example1')
            self.assertEqual(len(runtime.filtered_projects), 1)
            project = runtime.filtered_projects[0]
            self.assertEqual(project.path, path)

        for hostname, path in [
                ('example_machine', pathlib.Path('$INGIT_TEST_REPOS_PATH', 'example2')),
                ('special_machine', pathlib.Path('/special_path'))]:
            runtime = Runtime(self.runtime_config_path, self.repos_config_path, hostname=hostname)
            runtime.filter_projects(lambda name, *_: name == 'example2')
            self.assertEqual(len(runtime.filtered_projects), 1)
            project = runtime.filtered_projects[0]
            self.assertEqual(project.path, path)



"""JSON-based configuration I/O."""

import json
import json.decoder
import logging
import pathlib
import platform

from boilerplates.config import CONFIGS_PATH, normalize_path

from ._version import VERSION
from .runtime_interface import ask

_LOG = logging.getLogger(__name__)

JSON_INDENT = 2

JSON_ENSURE_ASCII = False

_CONFIG_DIRECTORY = CONFIGS_PATH.joinpath('ingit')
RUNTIME_CONFIG_PATH = _CONFIG_DIRECTORY.joinpath('ingit_config.json')
REPOS_CONFIG_PATH = _CONFIG_DIRECTORY.joinpath('ingit_repos.json')
REPO_LISTS_DIRECTORY_NAME = 'repos.d'


def json_to_str(data: dict) -> str:
    assert isinstance(data, dict), type(data)
    return json.dumps(data, indent=JSON_INDENT, ensure_ascii=JSON_ENSURE_ASCII)


def str_to_json(text: str) -> dict:
    """Convert JSON string into an object."""
    try:
        return json.loads(text)
    except json.decoder.JSONDecodeError as err:
        lines = text.splitlines(keepends=True)
        raise ValueError(
            f'\n{"".join(lines[max(0, err.lineno - 10):err.lineno])}{"-" * err.colno}'
            f'\n{"".join(lines[err.lineno:min(err.lineno + 10, len(lines))])}') from err


def json_to_file(data: dict, path: pathlib.Path) -> None:
    """Save JSON object to a file."""
    assert isinstance(data, dict), type(data)
    assert isinstance(path, pathlib.Path), type(path)
    text = json_to_str(data)
    with normalize_path(path).open('w', encoding='utf-8') as json_file:
        json_file.write(text)
        json_file.write('\n')


def file_to_json(path: pathlib.Path) -> dict:
    """Create JSON object from a file."""
    assert isinstance(path, pathlib.Path), type(path)
    with normalize_path(path).open('r', encoding='utf-8') as json_file:
        text = json_file.read()
    try:
        data = str_to_json(text)
    except ValueError as err:
        raise ValueError(f'in file "{path}"') from err
    return data


def default_runtime_configuration():
    return {
        'description': 'ingit runtime configuration file',
        'ingit-version': VERSION,
        'machines': [default_machine_configuration()]}


def default_machine_configuration(name=None):
    return {'interactive': True,
            'name': platform.node() if name is None else name,
            'repos_path': '~'}


def default_repos_configuration():
    return {
        'description': 'ingit repositories configuration file',
        'ingit-version': VERSION,
        'repos': []}


CONFIG_TYPES = {
    'runtime': (default_runtime_configuration, 'Runtime'),
    'repos': (default_repos_configuration, 'Repositories')}


def acquire_configuration(path: pathlib.Path, config_type: str):
    """Read (or create default) and return ingit configuration."""
    default_generator, config_type_name = CONFIG_TYPES[config_type]
    path = normalize_path(path)
    try:
        return file_to_json(path)
    except FileNotFoundError as err:
        ans = ask(f'{config_type_name} configuration file {path} does not exist.'
                  ' Create a default one?')
        if ans != 'y':
            raise err
        path.parent.mkdir(parents=True, exist_ok=True)
        config = default_generator()
        json_to_file(config, path)
        return config


def acquire_repos_configuration(path: pathlib.Path):
    """Read (or create default) and return ingit repositories configuration."""
    repos_config = acquire_configuration(path, 'repos')
    repo_lists_directory = normalize_path(path).parent.joinpath(REPO_LISTS_DIRECTORY_NAME)
    if not repo_lists_directory.is_dir():
        _LOG.debug('repos lists directory %s does not exist', repo_lists_directory)
        return repos_config

    _LOG.debug('loading repository lists from %s', repo_lists_directory)
    for list_path in repo_lists_directory.iterdir():
        if list_path.suffix != '.json':
            continue
        _LOG.debug('reading repos from %s', list_path)
        repos_config['repos'] += acquire_configuration(list_path, 'repos')['repos']
    return repos_config



"""Unit tests for ingit-only commands of ingit."""

import os
import pathlib
import tempfile
import unittest
import unittest.mock

from boilerplates.config import normalize_path
import git
import readchar

from ingit.json_config import json_to_file, acquire_configuration
from ingit.main import main

EXAMPLE_PROJECTS_PATH = pathlib.Path(os.environ.get('EXAMPLE_PROJECTS_PATH', '..'))
EXAMPLE_REPO_NAMES = ['argunparse', 'transpyle', 'typed-astunparse']
EXAMPLE_REPO_NAME = EXAMPLE_REPO_NAMES[0]


class Tests(unittest.TestCase):

    def setUp(self):
        for example_repo_name in EXAMPLE_REPO_NAMES:
            self.assertTrue(EXAMPLE_PROJECTS_PATH.joinpath(example_repo_name).is_dir())
        with tempfile.NamedTemporaryFile() as tmp_file:
            self.runtime_config_path = pathlib.Path(tmp_file.name)
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            self.runtime_config = acquire_configuration(self.runtime_config_path, 'runtime')
        with tempfile.NamedTemporaryFile() as tmp_file:
            self.repos_config_path = pathlib.Path(tmp_file.name)
        with unittest.mock.patch.object(readchar, 'readchar', return_value='y'):
            self.repos_config = acquire_configuration(self.repos_config_path, 'repos')

    def tearDown(self):
        self.repos_config_path.unlink()
        self.runtime_config_path.unlink()

    def call_main(self, *args, answer='y'):
        with unittest.mock.patch.object(readchar, 'readchar', return_value=answer):
            main(['--config', str(self.runtime_config_path), '--repos', str(self.repos_config_path)]
                 + list(args))

    def test_summary(self):
        for example_repo_name in EXAMPLE_REPO_NAMES:
            self.call_main('register', str(EXAMPLE_PROJECTS_PATH.joinpath(example_repo_name)))
            self.call_main('summary')

    def test_summary_empty(self):
        self.call_main('summary')

    def test_register(self):
        for i, example_repo_name in enumerate(EXAMPLE_REPO_NAMES):
            self.assertEqual(len(self.repos_config['repos']), i)
            self.call_main('register', str(EXAMPLE_PROJECTS_PATH.joinpath(example_repo_name)))
            self.repos_config = acquire_configuration(self.repos_config_path, 'repos')
            self.assertEqual(len(self.repos_config['repos']), i + 1)

    def test_repos_path_not_set(self):
        self.runtime_config['machines'][0]['repos_path'] = None
        json_to_file(self.runtime_config, self.runtime_config_path)
        self.runtime_config = acquire_configuration(self.runtime_config_path, 'runtime')

        repo_path = EXAMPLE_PROJECTS_PATH.joinpath(EXAMPLE_REPO_NAME)
        self.assertEqual(len(self.repos_config['repos']), 0)
        self.call_main('register', str(repo_path))
        self.repos_config = acquire_configuration(self.repos_config_path, 'repos')
        self.assertEqual(len(self.repos_config['repos']), 1)
        self.assertEqual(str(repo_path.resolve()), self.repos_config['repos'][0].get('path'))

        self.call_main('summary')

    def test_register_with_tags(self):
        for i, example_repo_name in enumerate(EXAMPLE_REPO_NAMES):
            self.assertEqual(len(self.repos_config['repos']), i)
            self.call_main('register', str(EXAMPLE_PROJECTS_PATH.joinpath(example_repo_name)),
                           '--tags', *[f'tag{_}' for _ in range(1, i + 2)])
            self.repos_config = acquire_configuration(self.repos_config_path, 'repos')
            self.assertEqual(len(self.repos_config['repos']), i + 1)

    def test_register_equivalents(self):
        path1 = pathlib.Path('.')
        path2 = pathlib.Path('.').resolve()
        path3 = pathlib.Path('..', pathlib.Path(os.getcwd()).name)
        self.call_main('register')
        with self.assertRaises(ValueError):
            self.call_main('register', str(path1))
        with self.assertRaises(ValueError):
            self.call_main('register', str(path2))
        with self.assertRaises(ValueError):
            self.call_main('register', str(path3))

    def test_register_outside(self):
        self.runtime_config['machines'][0]['repos_root'] = str(tempfile.gettempprefix())
        json_to_file(self.runtime_config, self.runtime_config_path)
        # self.runtime_config = acquire_configuration(self.runtime_config_path, 'runtime')
        repo_path = EXAMPLE_PROJECTS_PATH.joinpath(EXAMPLE_REPO_NAME)
        self.call_main('register', str(repo_path))

    def test_register_from_elsewhere(self):
        cwd = pathlib.Path(os.getcwd())
        with tempfile.TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            repo_path = cwd.resolve()
            self.call_main('register', str(repo_path))
            os.chdir(str(cwd))
        self.repos_config = acquire_configuration(self.repos_config_path, 'repos')
        self.assertEqual(len(self.repos_config['repos']), 1)
        repo_config = self.repos_config['repos'][0]
        repos_path = normalize_path(self.runtime_config['machines'][0]['repos_path'])
        self.assertEqual(repo_path.resolve(), pathlib.Path(repos_path, repo_config['path']))

    def test_register_nonrepo(self):
        repo_path = pathlib.Path('blah blah blah')
        with self.assertRaises(git.NoSuchPathError):
            self.call_main('register', str(repo_path))
        repo_path = pathlib.Path('test', 'examples')
        with self.assertRaises(git.InvalidGitRepositoryError):
            self.call_main('register', str(repo_path))



"""Monitoring progress of git operations."""

import functools
import operator
import logging
import shutil
import sys
import typing as t

import git
import git.remote

# list used by __create_operation_strings()
_KNOWN_OPERATIONS_STRINGS = {
    git.remote.RemoteProgress.BEGIN: 'begun',  # 1
    git.remote.RemoteProgress.END: 'ended',  # 2
    git.remote.RemoteProgress.COUNTING: 'counting objects',  # 4
    git.remote.RemoteProgress.COMPRESSING: 'compressing',  # 8
    git.remote.RemoteProgress.WRITING: 'writing',  # 16
    git.remote.RemoteProgress.RECEIVING: 'receiving objects',  # 32
    git.remote.RemoteProgress.RESOLVING: 'resolving deltas',  # 64
    git.remote.RemoteProgress.FINDING_SOURCES: 'finding sources',  # 128
    256: 'cloning url',
    512: 'fetching remote of submodule',
    # git.remote.RemoteProgress.CHECKING_OUT
}
_KNOWN_OPERATIONS: int = functools.reduce(operator.or_, _KNOWN_OPERATIONS_STRINGS.keys())
_KNOWN_OPERATIONS_PHASES = git.remote.RemoteProgress.BEGIN | git.remote.RemoteProgress.END

_CARET_UP = '\033[1A'

_LOG = logging.getLogger(__name__)


def _create_operation_strings(op_code: int):
    """Create operation strings."""
    operation_strings = []
    for key, value in _KNOWN_OPERATIONS_STRINGS.items():
        if op_code & key:
            operation_strings.append(value)
    unknown_operations = (
        op_code & ~_KNOWN_OPERATIONS)  # pylint: disable = invalid-unary-operand-type
    if unknown_operations:
        operation_strings.append(
            f'unknown operation code(s): {unknown_operations} ({unknown_operations:032b})')
    return operation_strings


class ActionProgress(git.remote.RemoteProgress):
    """Emulate usual git progress reports in the console when working with GitPython."""

    def __init__(self, inline: bool = True, f_d=None):
        """When no redirected_output_fd is given, use stdout."""
        assert isinstance(inline, bool)

        super().__init__()
        self.printed_lines = False
        self.line_len = 0  # type: int
        try:
            term_size = shutil.get_terminal_size()
        except ValueError:
            self.line_width = 100
        else:
            _LOG.log(logging.WARNING if term_size.columns < 16 else logging.NOTSET,
                     'detected terminal width is %i', term_size.columns)
            self.line_width = term_size.columns if term_size.columns else 100

        assert self.line_width > 0

        self.inline = inline
        self.f_d = f_d if f_d else sys.stdout

    def _print_without_nl(self, text: str):
        if self.inline:
            print(text, end='', flush=True, file=self.f_d)
        else:
            self._print_with_nl(text)

    def _print_with_nl(self, text: str):
        print(text, file=self.f_d)

    def update(self, op_code: int, cur_count: t.Any, max_count: t.Any = None, message: str = ''):
        """Override git.remote.RemoteProgress.update."""
        operation_strings = _create_operation_strings(op_code)
        description = f'{" ".join(operation_strings)}: ' if operation_strings else ''
        progress = ''
        if cur_count:
            progress += str(int(cur_count))
        if cur_count and max_count:
            progress += '/'
        if max_count:
            progress += str(int(max_count))
        if cur_count and max_count:
            max_c = max_count or 100
            progress = f'{cur_count / max_c:3.0%} ({progress})'
        if message:
            if message.startswith(', '):
                message = message[2:]
            message = ' - ' + message
        else:
            message = ''
        if self.inline and self.line_len > 0:
            if self.line_len < self.line_width:
                self._print_without_nl(f'\r{" " * self.line_len}\r')
            else:
                self._print_without_nl(f'\r{" " * (self.line_width - 1)}\r')
                self._print_without_nl(_CARET_UP)
                self._print_without_nl(f'{" " * (self.line_width - 1)}\r')
        line = f'{description}{progress}{message}'
        self.line_len = len(line)
        self._print_without_nl(line)
        self.printed_lines = True

    def finalize(self):
        """To be ran after the last progress report."""
        if self.printed_lines:
            if self.inline:
                self._print_with_nl('')
            self.printed_lines = False
            self.line_len = 0



"""Git repository data gathering and validation."""

import collections
import logging
import pathlib
import typing as t

import git

_LOG = logging.getLogger(__name__)


TrackingBranchTuple = t.NamedTuple('TrackingBranchTuple', [
    ('remote_name', str),
    ('tracking_branch_name', str),
    ('reference', git.Reference),
])


class RepoData:
    """Gather and validate data of a git repository."""

    def __init__(self, repo: git.Repo):
        assert isinstance(repo, git.Repo), type(repo)
        self._repo = repo
        self.remotes: t.Mapping[str, git.Remote] = {}
        self.branches: t.Mapping[str, git.Reference] = {}
        self._active_branch: t.Optional[str] = None
        self.remote_branches: t.Mapping[t.Tuple[str, str], git.Reference] = {}
        self.tracking_branches: t.Mapping[str, t.Optional[TrackingBranchTuple]] = {}

    @property
    def active_branch(self) -> t.Optional[str]:
        return self._active_branch

    @active_branch.setter
    def active_branch(self, branch_name: t.Optional[str]) -> None:
        self._active_branch = branch_name
        if branch_name is None:
            return
        if self._repo.active_branch.tracking_branch() is None:
            _LOG.warning('current branch "%s" in %s has no tracking branch',
                         self._active_branch, self._repo)

    @property
    def current_tracking_branch(self) -> t.Optional[str]:
        """Get the tracking branch of the current branch, if any."""
        if self._active_branch is None:
            return None
        tracking_branch = self.tracking_branches[self._active_branch]
        if tracking_branch is None:
            return None
        return f'{tracking_branch.remote_name}/{tracking_branch.tracking_branch_name}'

    def _refresh_tracking_branches(self) -> None:
        _tracking_branches: t.Dict[str, t.Optional[TrackingBranchTuple]] = {}
        for name, ref in self.branches.items():
            _tracking_branch = ref.tracking_branch()
            if _tracking_branch is None:
                _tracking_branches[name] = None
                continue
            assert isinstance(_tracking_branch, git.Reference), type(_tracking_branch)
            remote_name, _, branch_name = str(_tracking_branch).partition('/')
            _tracking_branches[name] = TrackingBranchTuple(
                remote_name, branch_name, _tracking_branch)
        self.tracking_branches = _tracking_branches

    @property
    def default_remote(self) -> t.Optional[str]:
        """Get the default remote name.

        Default remote name would be the remote name of the tracking branch of the current branch.
        It's None if the current branch has no tracking branch, or if there's no current branch.
        """
        if self.current_tracking_branch is None:
            return None
        assert self._active_branch is not None
        tracking_branch = self.tracking_branches[self._active_branch]
        assert tracking_branch is not None
        return tracking_branch.remote_name

    def _find_default_remote_ref(self) -> t.Optional[git.Remote]:
        default_remote = self.default_remote
        if default_remote is None:
            return None
        matching_remotes = [_ for _ in self._repo.remotes if str(_) == default_remote]
        assert len(matching_remotes) == 1, matching_remotes
        return matching_remotes[0]

    def _refresh_remotes(self) -> None:
        default_remote = self.default_remote
        default_remote_ref = self._find_default_remote_ref()
        if default_remote is None:
            self.remotes = collections.OrderedDict([])
        else:
            assert default_remote_ref is not None, default_remote
            self.remotes = collections.OrderedDict([(default_remote, default_remote_ref)])
        self.remotes.update(collections.OrderedDict([(str(_), _) for _ in self._repo.remotes]))

    def refresh(self) -> None:
        """Refresh repository data."""
        self.active_branch = None
        if self._repo.branches:
            try:
                self.active_branch = str(self._repo.active_branch)
            except TypeError:
                _LOG.warning('repository %s is not on any branch', self._repo)

        self.branches = collections.OrderedDict(
            [] if self._active_branch is None
            else [(self._active_branch, self._repo.active_branch)])
        self.branches.update(collections.OrderedDict([(str(_), _) for _ in self._repo.branches]))
        assert all(_ is not None for name, _ in self.branches.items()), (self._repo, self.branches)
        self._refresh_tracking_branches()
        self._refresh_remotes()

        self.remote_branches = {
            (remote_name, str(_).replace(f'{remote_name}/', '')): _
            for remote_name, remote in self.remotes.items() for _ in remote.refs}

    def generate_repo_configuration(self) -> t.Dict[str, t.Any]:
        assert self._repo.working_tree_dir is not None
        path = pathlib.Path(self._repo.working_tree_dir)
        return {
            'name': path.name,
            'path': str(path.resolve()),
            'remotes': collections.OrderedDict(
                [(name, remote.url) for name, remote in self.remotes.items()]),
            'tags': []}

    def __str__(self):
        return f'{self._repo}:(branches={self.tracking_branches},remotes={self.remotes})'

    @property
    def tags(self):
        return self._repo.tags

    @property
    def git(self):
        return self._repo.git



"""Setup script for ingit package."""

import boilerplates.setup


class Package(boilerplates.setup.Package):
    """Package metadata."""

    name = 'ingit'
    description = 'Tool for managing a large collection of repositories in git.'
    url = 'https://github.com/mbdevpl/ingit'
    license_str = 'GPL-3.0-or-later'
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Version Control',
        'Topic :: Software Development :: Version Control :: Git',
        'Topic :: System :: Archiving :: Backup',
        'Topic :: System :: Archiving :: Mirroring',
        'Topic :: System :: Monitoring',
        'Topic :: Utilities',
        'Typing :: Typed']
    keywords = ['tools', 'vcs', 'repository management', 'git', 'submodules']
    entry_points = {'console_scripts': ['ingit = ingit.__main__:main']}


if __name__ == '__main__':
    Package.setup()



"""Command-line interface of ingit."""

import argparse
import logging
import pathlib
import sys

import argcomplete
from boilerplates.cli import \
    ArgumentDefaultsAndRawDescriptionHelpFormatter, make_copyright_notice, add_version_option, \
    add_verbosity_group, get_logging_level, dedent_except_first_line
from boilerplates.config import initialize_config_directory

from ._version import VERSION
from .json_config import RUNTIME_CONFIG_PATH, REPOS_CONFIG_PATH
from .runtime import Runtime

_LOG = logging.getLogger(__name__)

PREDICATE_EXAMPLES = ['''name.startswith('py_')''', ''' 'python' in tags''']

REGEX_EXAMPLES = ['^py_.*', '^python$']

SUGGESTED_TAGS = [
    'appveyor', 'archived', 'assembla', 'bash', 'bitbucket', 'c', 'c++', 'c#', 'css', 'cython',
    'docker', 'fortran', 'gist', 'github', 'html', 'java', 'javascript', 'latex', 'opencl', 'php',
    'python', 'ruby', 'travis', 'vsonline']

OUT = logging.getLogger('ingit.interface.print')


def prepare_parser():
    """Prepare command-line arguments parser."""
    parser = argparse.ArgumentParser(
        prog='ingit',
        description='''Tool for managing a large collection of repositories in git. If you have
        100 git-versioned projects, keeping tabs on everything can be quite troublesome.''',
        epilog=make_copyright_notice(
            2015, 2026, license_name='GNU General Public License v3 or later (GPLv3+)',
            url='https://github.com/mbdevpl/ingit'),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=True)
    add_version_option(parser, VERSION)

    interactivity_group = parser.add_mutually_exclusive_group(required=False)
    interactivity_group.add_argument(
        '--batch', '--non-interactive', dest='batch', action='store_true',
        help='run ingit in non-interactive mode')
    interactivity_group.add_argument(
        '--interactive', dest='batch', action='store_false',
        help='force interactive mode even if configuration sets batch mode as default')
    parser.set_defaults(batch=None)

    add_verbosity_group(parser)

    parser.add_argument(
        '--config', metavar='PATH', type=str, default=str(RUNTIME_CONFIG_PATH),
        help='''path to the runtime configuration file;
        can be absolute, or relative to current woking directory''')

    parser.add_argument(
        '--repos', metavar='PATH', type=str, default=str(REPOS_CONFIG_PATH),
        help='''path to the projects list file;
        can be absolute, or relative to current woking directory''')

    parser.add_argument(
        '--predicate', '-p', type=str, default=None, help=f'''a Python expression used to select
        repositories operated on; it is evaluated on each repository metadata;
        examples: "{'", "'.join(PREDICATE_EXAMPLES)}"''')
    parser.add_argument(
        '--regex', '-r', type=str, default=None, help=f'''a regular expression used to select
        repositories operated on; repository matches if any of its metadata match;
        examples: "{'", "'.join(REGEX_EXAMPLES)}"''')

    commands = {
        'summary': (
            'show summary of registered repositories and status of configured repository root',
            '''First of all, print a list of registered repositories. By default, all
            registered repositories are listed, but, as in case of most commands, the
            results can be filtered via a predicate or regex.

            Independently, print a list of all unregistered repositories and all not
            versioned paths present in the configured repositories root.'''),
        'register': (
            'start tracking a repository in ingit',
            '''The initial configuration is set according to basic repository information:
            its root directory name becomes "name" and its currently configured remotes
            become "remotes". You can edit the configuration manually afterwards.

            The final "path" to the repository stored in the configuration depends on the
            "repos_path" in runtime configuration. The configured "path" will be:

            *   resolved absolute path if there is no "repos_path" configured or
                repository path is outside of the "repos_path";
            *   resolved relative path to the "repos_path", if the repository path is
                within it;
            *   nothing (i.e. not stored) if the if the repository is stored directly in
                "repos_path" (i.e. there are no intermediate directories).

            Behaviour of storing relative/no paths in some cases is implemented to make
            configuration file much less verbose in typical usage scenarios. To prevent
            this behaviour, and force all repository paths to be absolute, simply set the
            "repos_path" in your runtime configuration to JSON "null".'''),
        'foreach': (
            'execute a custom command',
            '''The given command is executed in a shell in working directory of each
            project.'''),
        'clone': (
            'perform git clone',
            '''Execute "git clone <remote-url> --recursive --origin <remote-name> <path>",
            where values of <path> and <remote-...> are taken from default remote
            configuration of the repository.

            After cloning, add all remaining configured remotes to the repository and
            fetch them.'''),
        'init': (
            'perform git init',
            '''Execute "git init", followed by "git remote add" for each configured
            remote.'''),
        'fetch': (
            'perform git fetch',
            '''Execute "git fetch <remote-name>", where the remote name is the remote of
            the current tracking branch, or all remotes of the repository if there's no
            tracking branch, or repository is in detached head state.'''),
        'checkout': (
            'perform git checkout',
            '''Interactively select revision to checkout from list of local branches, remote
            non-tracking branches and local tags.

            The list of branches to select from is composed by combining:

            *   local branches
            *   non-tracking branches on all remotes
            *   local tags

            Checking out a remote branch will create a local branch with the same unless
            it already exists. If it already exists, repository will end up in detached
            head state.

            Also, checking out any tag will put repository in a detached head state.'''),
        'merge': (
            'perform git merge (not yet implemented)',
            '''Not yet implemented! The following functionality is intended.

            Interactively merge all branches to their tracking branches. For each not
            merged <branch>-<tracking-branch> pair, execute
            "git checkout <branch>" and then if the merge is fast-forward,
            automatically execute "git merge <tracking-branch> --ff-only". If not, then
            show more information about the situation of the repository, and propose:

            *   "git merge --log <tracking-branch>",
            *   "git rebase -i <tracking-branch>" and
            *   "git reset --hard <tracking-branch>".

            If repository is dirty when this command is executed, do nothing. After work
            is done, return to the originally checked-out branch.'''),
        'push': (
            'perform git push (not yet fully implemented)',
            '''Execute "git push <remote-name> <branch>:<tracking-branch-name>" for the
            active branch.'''),
        'gc': ('perform git gc', 'Execute "git gc --agressive --prune".'),
        'status': (
            'perform git status, as well as other diagnostic git commands',
            '''Perform git status, as well as other diagnostic git commands.

            Execute:

            *   "git status --short --branch" to inform about any uncommitted changes,
            *   "git log tracking_branch..branch" to inform about commits that are not
                yet pushed to the remote,
            *   "git log branch..tracking_branch" to inform about commits that are not
                yet merged from the remote.

            Additionally, compare registered remotes with actual remotes to make sure
            that ingit configuration is in sync with the repository metadata..''')}

    subparsers = parser.add_subparsers(
        dest='command', metavar='command', help=f'''main command to execute; one of:
        "{'", "'.join(commands.keys())}";
        run "ingit command --help" to see detailed help for a given command''')

    _prepare_command_subparsers(subparsers, commands)
    argcomplete.autocomplete(parser)

    return parser


def _prepare_command_subparsers(subparsers, commands):
    for command, (help_, description) in commands.items():
        subparser = subparsers.add_parser(
            command, help=help_, formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)
        subparser.description = dedent_except_first_line(description)
        if command == 'register':
            subparser.add_argument(
                '--tags', metavar='TAG', type=str, default=None, nargs='+',
                help='set tags for this repository, they will be added to initial configuration') \
                .completer = argcomplete.completers.ChoicesCompleter(choices=SUGGESTED_TAGS)
            subparser.add_argument(
                'path', metavar='PATH', type=str, nargs='?',
                help='''path to root directory of repository, use current working directory
                if not provided''')
        elif command == 'foreach':
            subparser.add_argument(
                'cmd', metavar='COMMAND', type=str,
                help='command to be executed in shell in working directory of each project')
            # subparser.add_argument(
            #     '--recursive', action='store_true',
            #     help='(not yet implemented)')
            subparser.add_argument(
                '--timeout', metavar='SECONDS', type=float, default=None,
                help='timeout of the command (in seconds)')
        elif command == 'fetch':
            subparser.add_argument(
                '--all', action='store_true',
                help='fetch all remotes in all cases')
        elif command == 'push':
            subparser.add_argument(
                '--all', action='store_true',
                help='''(not yet implemented) execute the push for every branch that has a remote
                tracking branch''')
        elif command == 'status':
            subparser.add_argument(
                '-i', '--ignored', action='store_true',
                help='''include ignored files in the status report
                (identical to "--ignored" flag on "git status")''')


def _prepare_command_options(command, parsed_args):
    command_options = {}
    if command == 'register':
        command_options['tags'] = parsed_args.tags
        command_options['path'] = pathlib.Path(
            '.' if parsed_args.path is None else parsed_args.path)
    elif command == 'foreach':
        command_options['cmd'] = parsed_args.cmd
        command_options['timeout'] = parsed_args.timeout
    elif command == 'fetch':
        command_options['all_remotes'] = parsed_args.all
    elif command == 'push':
        command_options['all_branches'] = parsed_args.all
    elif command == 'status':
        command_options['ignored'] = parsed_args.ignored
    return command_options


def main(args=None):
    """Parse command line arguments and run ingit accordingly."""
    parser = prepare_parser()
    parsed_args = parser.parse_args(args)
    if (parsed_args.predicate is not None or parsed_args.regex is not None) \
            and parsed_args.command == 'register':
        parser.error(f'project filtering is not applicable to "{parsed_args.command}" command'
                     ' -- it can be used only with summary command and with git-like commands')

    try:
        level = get_logging_level(parsed_args)
    except ValueError as err:
        _LOG.debug('failed to get interface verbosity level', exc_info=True)
        OUT.critical('%s\ningit: error: %s', parser.format_usage(), err.args[0])
        sys.exit(1)
    OUT.setLevel(level)
    assert level in {OUT.getEffectiveLevel(), logging.NOTSET}, (level, OUT.getEffectiveLevel())

    OUT.info('parsed args: %s', parsed_args)

    initialize_config_directory('ingit')

    runtime_config_path = pathlib.Path(parsed_args.config)
    repos_config_path = pathlib.Path(parsed_args.repos)

    if parsed_args.predicate is None:
        predicate = None
    else:
        predicate_code = f"lambda name, tags, path, remotes: ({parsed_args.predicate})"
        _LOG.warning('prepared predicate lambda: %s', predicate_code)
        predicate = eval(predicate_code)  # pylint: disable = eval-used

    if parsed_args.regex is None:
        regex = None
    else:
        regex = parsed_args.regex

    command = parsed_args.command
    if command is None:
        parser.error('no command provided')

    command_options = _prepare_command_options(command, parsed_args)

    interactive = None if parsed_args.batch is None else not parsed_args.batch
    runtime = Runtime(runtime_config_path, repos_config_path, interactive=interactive)

    if predicate is not None:
        runtime.filter_projects(predicate)
    if regex is not None:
        runtime.filter_projects(regex)
    runtime.execute(command, **command_options)

    # run(runtime_config_path, repos_config_path, predicate, regex, command, **command_options)
